from copy import deepcopy
from typing import List, Callable, Tuple, Dict, Optional, Set

from kazu.data.data import Entity, CharSpan

import spacy


def _copy_ent_with_new_spans(
    entity: Entity,
    spans: List[Tuple[int, int]],
    text: str,
    join_str: str,
    rule_name: Optional[str] = None,
):
    attrb_dict = deepcopy(entity).__dict__
    attrb_dict.pop("spans")
    attrb_dict.pop("match_norm")
    attrb_dict.pop("match")
    attrb_dict.pop("start")
    attrb_dict.pop("end")
    ent = Entity.from_spans(spans=spans, text=text, join_str=join_str, **attrb_dict)
    if rule_name:
        ent.metadata["split_rule"] = rule_name
    return ent


class SplitOnConjunctionPattern:
    def __init__(self, spacy_pipeline: spacy.Language):
        """
        analyse

        :param pattern:
        """
        self.nlp = spacy_pipeline

    def __call__(self, entity: Entity, text: str) -> List[Entity]:
        doc = self.nlp(entity.match)
        if any((x in entity.match for x in [" and ", " or ", " nor "])):
            return self.run_conjunction_rules(doc, entity, text)
        else:
            return []

    def run_conjunction_rules(self, doc, entity, text):
        ents = []
        noun_chunks = list(doc.noun_chunks)
        if len(noun_chunks) > 0:
            anchor_chunk = noun_chunks[-1]
            anchor = None
            for tok in anchor_chunk:
                if tok.dep_ == "conj":
                    anchor = tok
            if anchor is not None:
                ents.append(
                    _copy_ent_with_new_spans(
                        spans=[
                            (
                                entity.start + anchor_chunk.start_char,
                                entity.start + anchor_chunk.end_char,
                            )
                        ],
                        text=text,
                        entity=entity,
                        join_str=" ",
                        rule_name="spacy noun chunk tags",
                    )
                )
                for chunk in noun_chunks[:-1]:
                    if anchor in chunk.conjuncts:
                        ents.append(
                            _copy_ent_with_new_spans(
                                spans=[
                                    (
                                        entity.start + chunk.start_char,
                                        entity.start
                                        + (chunk.end_char - chunk.start_char)
                                        + chunk.start_char,
                                    ),
                                    (
                                        entity.start + anchor.idx,
                                        entity.start + len(anchor) + anchor.idx,
                                    ),
                                ],
                                text=text,
                                entity=entity,
                                join_str=" ",
                                rule_name="spacy noun chunk tags",
                            )
                        )

        return ents


class SplitOnNumericalListPatternWithPrefix:
    """
    split a string of numerically incrementing parts:

    .. testsetup::

        from kazu.steps.ner.entity_post_processing import SplitOnNumericalListPatternWithPrefix
        from kazu.data.data import Entity

    .. testcode::

        splitter = SplitOnNumericalListPatternWithPrefix()
        ent = Entity.load_contiguous_entity(
            start=0, end=8, namespace="test", entity_class="gene", match="BRCA1/2/3"
        )
        print(splitter(ent, "BRCA1/2/3 are oncogenes"))

    .. testoutput::

        [BRCA1:gene:test:0:5, BRCA2:gene:test:0:7, BRCA3:gene:test:0:9]


    """

    def __init__(self, pattern: str = "/"):
        """
        pattern to split the string on (typically "/")

        :param pattern:
        """
        self.pattern = pattern

    def __call__(self, entity: Entity, text: str) -> List[Entity]:
        input_ent_start = entity.start
        parts = entity.match.split(self.pattern)
        new_ents = []
        if len(parts) > 1:
            i = None
            for i, char in enumerate(reversed(list(parts[0]))):
                if not char.isdigit():
                    break

            new_ents.append(
                _copy_ent_with_new_spans(
                    entity=entity,
                    spans=[(input_ent_start, input_ent_start + len(parts[0]))],
                    text=text,
                    join_str="",
                    rule_name=self.__class__.__name__,
                )
            )
            if i:
                prefix = parts[0][: len(parts[0]) - i]
                span_offset = len(parts[0]) + len(self.pattern)
                for part in parts[1:]:
                    if not part.isdigit():
                        continue
                    else:
                        new_spans = [
                            (input_ent_start, input_ent_start + len(prefix)),
                            (
                                input_ent_start + span_offset,
                                input_ent_start + span_offset + len(part),
                            ),
                        ]
                        span_offset += len(self.pattern) + len(part)
                        new_ents.append(
                            _copy_ent_with_new_spans(
                                entity=entity,
                                spans=new_spans,
                                text=text,
                                join_str="",
                                rule_name=self.__class__.__name__,
                            )
                        )
        return new_ents


class NonContiguousEntitySplitter:
    """
    Some simple rules to split non-contiguous entities into component entities
    """

    def __init__(self, entity_conditions: Dict[str, List[Callable[[Entity, str], List[Entity]]]]):

        self.entity_conditions = entity_conditions

    def __call__(self, entity: Entity, text: str) -> List[Entity]:
        existing_offsets: Set[CharSpan] = set()
        existing_offsets.update(entity.spans)

        new_ents = []
        for rule in self.entity_conditions.get(entity.entity_class, []):
            found_ents = rule(entity, text)
            for found_ent in found_ents:
                # only add new ent if offsets have changed
                if any(span not in existing_offsets for span in found_ent.spans):
                    new_ents.append(found_ent)
                    existing_offsets.update(found_ent.spans)
        return new_ents



import logging
import re
from abc import ABC, abstractmethod
from copy import deepcopy
from dataclasses import dataclass, field
from typing import List, Tuple, Dict, Optional, Set

import torch
from kazu.data.data import Entity, ENTITY_OUTSIDE_SYMBOL, ENTITY_START_SYMBOL, IS_SUBSPAN
from torch import Tensor

logger = logging.getLogger(__name__)


@dataclass
class TokenizedWord:
    """
    A convenient container for a word, which may be split into multiple tokens by e.g. WordPiece tokenisation
    """

    token_ids: List[int]
    #: token string representations
    tokens: List[str]
    #: tensor of token logit softmax
    token_confidences: Tensor
    #: character indices of tokens
    token_offsets: List[Tuple[int, int]]
    #: char start index of word
    word_char_start: int
    #: char end index of word
    word_char_end: int
    word_id: int


@dataclass
class TokWordSpan:
    """
    dataclass for a span (i.e. a List[TokenizedWord] representing an NE)
    """

    #: entity_class
    clazz: str
    #: is this a subspan? only relevant if using SmartSpanFinder
    subspan: Optional[bool] = None
    #: words associated with this span
    tok_words: List[TokenizedWord] = field(default_factory=list)


class SpanFinder(ABC):
    """
    Since Bert like models use wordpiece tokenisers to handle the OOV problem, we need to reconstitute this info
    back into document character indices to represent actual NEs.

    Since transformer NER is an inprecise art, we may want to use different logic in how this works, so we can
    subclass this class to determine how this should be done

    The __call__ method of this class operates on a list of TokenizedWord, processing each sequentially according to
    a logic determined by the implementing class.

    After being called, the spans can be accessed via self.closed_spans. This is a list of TokWordSpan.
    """

    def __init__(self, text: str, id2label: Dict[int, str]):
        """

        :param text: the raw text to be processed
        :param id2label: BIO to class label mappings
        """
        self.text = text
        self.active_spans: List[TokWordSpan] = []
        self.words: List[TokenizedWord] = []
        self.span_breaking_chars = {"(", ")", ";"}
        self.closed_spans: List[TokWordSpan] = []
        self.id2label = id2label

    @abstractmethod
    def get_bio_and_class_labels(self, word: TokenizedWord) -> Set[Tuple[str, Optional[str]]]:
        """
        return a set of Tuple[<BIO label>,Optional[<class label>]] for a TokenizedWord. Optional[<class label>] is None
        if the BIO label is "O".

        :param word:
        :return:
        """
        pass

    @abstractmethod
    def span_continue_condition(
        self, word: TokenizedWord, bio_and_class_labels: Set[Tuple[str, Optional[str]]]
    ):
        """
        based upon some logic, determine whether a span should continue or not

        :param word:
        :param bio_and_class_labels:
        :return:
        """
        pass

    def _update_active_spans(
        self, bio_and_class_labels: Set[Tuple[str, Optional[str]]], word: TokenizedWord
    ):
        """
        updates any active spans. If a B label is detected in an active span, make a copy and add to closed spans,
        as it's likely the start of another entity of the same class (but we still want to keep the original span open)

        :param bio_and_class_labels: BIO and optional class label set
        :param word:
        :return:
        """
        for span in self.active_spans:
            for bio, c in bio_and_class_labels:
                if bio == ENTITY_START_SYMBOL and c == span.clazz:
                    # a new word of the same class is beginning, so add a copy of the currently open TokWordSpan to the
                    # closed spans
                    self.closed_spans.append(deepcopy(span))
                if c == span.clazz:
                    # the word is the same class as the active TokWordSpan, so extend it
                    span.tok_words.append(word)

    def start_span(
        self,
        bio_and_class_labels: Set[Tuple[str, Optional[str]]],
        word: TokenizedWord,
        subspan: Optional[bool],
    ):
        """
        start a new TokWordSpan if a B label is detected

        :param bio_and_class_labels:
        :param word:
        :return:
        """
        for bio, clazz in bio_and_class_labels:
            if bio == ENTITY_START_SYMBOL and clazz is not None:
                span = TokWordSpan(clazz=clazz, subspan=subspan, tok_words=[word])
                self.active_spans.append(span)

    def close_spans(self):
        """
        close any active spans
        """
        for active_span in self.active_spans:
            if len(active_span.tok_words) > 0:
                self.closed_spans.append(active_span)
        self.active_spans = []

    @abstractmethod
    def process_next_word(self, word: TokenizedWord):
        """
        process the next word in the sequence, according to some logic

        :param word:
        :return:
        """
        pass

    def __call__(self, words: List[TokenizedWord]):
        for word in words:
            self.process_next_word(word)
        self.close_spans()


class SimpleSpanFinder(SpanFinder):
    """
    A basic implementation of SpanFinder. Uses highest confidence labels only
    """

    def __init__(self, text: str, id2label: Dict[int, str]):

        super().__init__(text, id2label)

    def get_bio_and_class_labels(self, word: TokenizedWord) -> Set[Tuple[str, Optional[str]]]:
        bio_and_class_labels: Set[Tuple[str, Optional[str]]] = set()
        most_conf_index_per_token = torch.argmax(word.token_confidences, dim=1)
        for confidence_index in most_conf_index_per_token:
            bio_label = self.id2label[confidence_index.item()]
            if bio_label == ENTITY_OUTSIDE_SYMBOL:
                bio_and_class_labels.add(
                    (
                        bio_label,
                        None,
                    )
                )
            else:
                bio_label, class_label = bio_label.split("-")
                bio_and_class_labels.add(
                    (
                        bio_label,
                        class_label,
                    )
                )

        return bio_and_class_labels

    def span_continue_condition(
        self, word: TokenizedWord, bio_and_class_labels: Set[Tuple[str, Optional[str]]]
    ):
        """
        A potential entity span will end if any of the following conditions are met:

        1. any of the BIO classes for word are O
        2. The previous character to the word is in the set of self.span_breaking_chars

        :param word:
        :return:
        """
        classes_set = set(x[1] for x in bio_and_class_labels)
        if None in classes_set or self.text[word.word_char_start - 1] in self.span_breaking_chars:
            return False
        else:
            return True

    def process_next_word(self, word: TokenizedWord):
        """
        process the next word in the sequence, updating span information accordingly

        :param word:
        :return:
        """

        bio_and_class_labels = self.get_bio_and_class_labels(word)
        if not self.words or len(self.active_spans) == 0:
            self.start_span(bio_and_class_labels=bio_and_class_labels, word=word, subspan=None)
        elif self.span_continue_condition(word, bio_and_class_labels):
            self._update_active_spans(bio_and_class_labels, word)
        else:
            self.close_spans()
            self.start_span(bio_and_class_labels=bio_and_class_labels, word=word, subspan=None)
        self.words.append(word)


class SmartSpanFinder(SpanFinder):
    """
    A more complicated implementation of SpanFinder using 'soft labels' - i.e. nested and overlapping entities of
    different classes may be produced, if the associated label is above some threshold
    """

    def __init__(self, threshold: float, text: str, id2label: Dict[int, str]):
        super().__init__(text, id2label)
        self.text = text
        self.threshold = threshold
        # we include a whitespace here so that new spans (potential entities) are started for each word
        self.span_breaking_chars = {"(", ")", ";", " "}

    def get_bio_and_class_labels(self, word: TokenizedWord) -> Set[Tuple[str, Optional[str]]]:
        """
        returns bio and class labels if their confidence is above the configured threshold

        :param word:
        :return:
        """
        bio_and_class_labels: Set[Tuple[str, Optional[str]]] = set()
        indices_of_confs_above_threshold = torch.argwhere(word.token_confidences > self.threshold)
        # the 0th dimension is the order of the tokens in the word, which we don't need.
        # The first dimension represents which label the confidence score relates to.
        label_indices_above_threshold = indices_of_confs_above_threshold[:, 1]
        for label_index in label_indices_above_threshold:
            bio_label = self.id2label[label_index.item()]
            if bio_label == ENTITY_OUTSIDE_SYMBOL:
                bio_and_class_labels.add(
                    (
                        bio_label,
                        None,
                    )
                )
            else:
                bio_label, class_label = bio_label.split("-")
                bio_and_class_labels.add(
                    (
                        bio_label,
                        class_label,
                    )
                )
        return bio_and_class_labels

    def span_continue_condition(
        self, word: TokenizedWord, bio_and_class_labels: Set[Tuple[str, Optional[str]]]
    ):
        """
        A potential entity span must continue if any of the following conditions are met:

        1. The previous character to the word is not in the set of self.span_breaking_chars
        2. There are any entity class assignments in any of the tokens in the TokenizedWord under consideration.

        :param word:
        :return:
        """
        classes_set = set(x[1] for x in bio_and_class_labels)
        classes_set.discard(None)
        if (
            self.text[word.word_char_start - 1] not in self.span_breaking_chars
            or len(classes_set) > 0
        ):
            return True
        else:
            return False

    def process_next_word(self, word: TokenizedWord):
        """
        process the next word in the sequence, updating span information accordingly

        :param word:
        :return:
        """

        bio_and_class_labels = self.get_bio_and_class_labels(word)
        if not self.words or len(self.active_spans) == 0:
            self.start_span(bio_and_class_labels=bio_and_class_labels, word=word, subspan=False)
        elif self.span_continue_condition(word, bio_and_class_labels):
            self._update_active_spans(bio_and_class_labels, word)
            # we start a new span here for each new word if B is detected as a soft label
            self.start_span(bio_and_class_labels=bio_and_class_labels, word=word, subspan=True)
        else:
            self.close_spans()
            self.start_span(bio_and_class_labels=bio_and_class_labels, word=word, subspan=False)

        self.words.append(word)


class TokenizedWordProcessor:
    """
    Because of the inherent obscurity of the inner workings of transformers, sometimes they produce BIO tags that
    don't correctly align to whole words, or maybe the classic BIO format gets confused by nested entities.

    This class is designed to work when an entire sequence of NER labels is known and therefore we can apply some
    post-processing logic. Namely, we use the SpanFinder class to find entity spans according to their internal logic
    """

    def __init__(
        self,
        confidence_threshold: Optional[float],
        id2label: Dict[int, str],
        detect_subspans: bool = False,
        strip_re: Optional[Dict[str, str]] = None,
    ):
        """

        :param confidence_threshold: optional threshold if using :class:`SmartSpanFinder`\\ . Ignored if detect_subspans is false
        :param id2label: mapping of label int id to str label
        :param detect_subspans: use :class:`SmartSpanFinder` if True. A confidence_threshold must be provided
        :param strip_re: an optional dict of {<entity_class>:<python regex to remove>} to process NER results that the
            model frequently misclassifies.
        """

        self.id2label = id2label
        self.detect_subspans = detect_subspans
        self.confidence_threshold = confidence_threshold
        self.strip_re = (
            {k: re.compile(v) for k, v in strip_re.items()} if strip_re is not None else None
        )

    def _make_smart_span_finder(self, text: str):
        logger.debug(
            "detect_subspans is %s. Using %s, confidence threshold: %s",
            self.detect_subspans,
            SmartSpanFinder.__class__.__name__,
            self.confidence_threshold,
        )
        if isinstance(self.confidence_threshold, float):
            return SmartSpanFinder(
                threshold=self.confidence_threshold,
                text=text,
                id2label=self.id2label,
            )
        else:
            raise ValueError(
                f"tried to instantiate {SmartSpanFinder.__class__.__name__} but confidence threshold is not a float"
            )

    def _make_simple_span_finder(self, text: str):
        logger.debug(
            "detect_subspans is %s. Using %s",
            self.detect_subspans,
            SimpleSpanFinder.__class__.__name__,
        )
        return SimpleSpanFinder(text=text, id2label=self.id2label)

    def make_span_finder(self, text: str) -> SpanFinder:
        if self.detect_subspans:
            return self._make_smart_span_finder(text)
        else:
            return self._make_simple_span_finder(text)

    def __call__(self, words: List[TokenizedWord], text: str, namespace: str) -> List[Entity]:
        span_finder: SpanFinder = self.make_span_finder(text)
        span_finder(words)
        ents = self.spans_to_entities(span_finder.closed_spans, text, namespace)
        return ents

    def spans_to_entities(
        self, spans: List[TokWordSpan], text: str, namespace: str
    ) -> List[Entity]:
        """
        convert spans to instances of Entity, adding in namespace info as appropriate

        :param spans: list of TokWordSpan to consider
        :param text: original text
        :param namespace: namespace to add to Entity
        :return:
        """
        entities = []
        for span in spans:
            start, end = self.calculate_span_offsets(span.tok_words)
            # sometimes the tokenizer seems to mess up the offsets
            match_str = text[start:end]
            if (
                len(match_str) > 1
                and match_str[-1] == " "
                and any(char.isalpha() for char in match_str)
            ):
                end = end - 1
            match_str = text[start:end]
            if span.subspan is not None:
                metadata = {IS_SUBSPAN: span.subspan}
            else:
                metadata = {}

            match_str, end = self.attempt_strip_suffixes(start, end, match_str, span.clazz)

            entity = Entity.load_contiguous_entity(
                start=start,
                end=end,
                match=match_str,
                namespace=namespace,
                entity_class=span.clazz,
                metadata=metadata,
            )
            entities.append(entity)

        return entities

    def calculate_span_offsets(self, words: List[TokenizedWord]) -> Tuple[int, int]:
        starts, ends = [], []
        for word in words:
            starts.append(word.word_char_start)
            ends.append(word.word_char_end)
        return min(starts), max(ends) + 1

    def attempt_strip_suffixes(
        self, start: int, end: int, match_str: str, clazz: str
    ) -> Tuple[str, int]:
        """
        transformers sometimes get confused about precise offsets, depending on the training data
        (e.g. "COX2" is often classified as "COX2 gene"). This method offers light post-processing
        to correct these, for better entity linking results

        :param start: original start
        :param end: original end
        :param match_str: original string
        :param clazz: entity class
        :return: new string, new end
        """
        if self.strip_re is not None:
            suffixes_re = self.strip_re.get(clazz)
            if suffixes_re is not None:
                match_str = suffixes_re.sub("", match_str)
                end = len(match_str) + start
        return match_str, end



from kazu.steps.step import Step, document_iterating_step, document_batch_step
from kazu.steps.document_post_processing.abbreviation_finder import AbbreviationFinderStep
from kazu.steps.joint_ner_and_linking.explosion import ExplosionStringMatchingStep
from kazu.steps.linking.dictionary import DictionaryEntityLinkingStep
from kazu.steps.linking.post_processing.mapping_step import MappingStep
from kazu.steps.ner.hf_token_classification import TransformersModelForTokenClassificationNerStep
from kazu.steps.ner.seth import SethStep
from kazu.steps.ner.spacy_ner import SpacyNerStep
from kazu.steps.other.cleanup import CleanupStep
from kazu.steps.other.merge_overlapping_ents import MergeOverlappingEntsStep
from kazu.steps.other.stanza import StanzaStep



from typing import Iterator, Dict, Any

from torch.utils.data import IterableDataset
from transformers import BatchEncoding


class HFDataset(IterableDataset):
    def __getitem__(self, index):
        pass  # type: ignore[empty-body]

    def __init__(self, encodings: BatchEncoding):
        """
        simple implementation of :class:`torch.utils.data.IterableDataset`\\ , producing HF tokenizer input_id

        :param encodings:
        """
        self.encodings = encodings
        self.dataset_size = len(encodings.data["input_ids"])

    def __iter__(self) -> Iterator[Dict[str, Any]]:

        for i in range(self.dataset_size):
            yield {
                "input_ids": self.encodings.data["input_ids"][i],
                "attention_mask": self.encodings.data["attention_mask"][i],
                "token_type_ids": self.encodings.data["token_type_ids"][i],
            }



from datetime import datetime
import logging
import os
import time
import uuid
from typing import List, Dict, Optional, Protocol, Iterable, Collection

import psutil
from hydra.utils import instantiate
from omegaconf import DictConfig

from kazu.data.data import Document, PROCESSING_EXCEPTION
from kazu.steps import Step
from kazu.utils.utils import PathLike, as_path

logger = logging.getLogger(__name__)


def load_steps_and_log_memory_usage(cfg: DictConfig) -> List[Step]:
    """
    Loads steps based on the pipeline config, and log the memory increase after loading each step.

    Note that you can instantiate the pipeline directly from the config in a way that gives the
    same results, but this is useful for monitoring/debugging high memory usage.

    :param cfg: An OmegaConf config object for the kazu :class:`.Pipeline`\\ .
        Normally created from hydra config files.
    :return: The instantiated steps from the pipeline config
    """
    steps = []
    for step_cfg in cfg.Pipeline.steps:
        prev_memory = psutil.Process(os.getpid()).memory_info().rss / 1024**2
        new_step = instantiate(step_cfg)
        steps.append(new_step)
        new_memory = psutil.Process(os.getpid()).memory_info().rss / 1024**2
        mem_increase = ((new_memory - prev_memory) / prev_memory) * 100

        logger.info(
            f"loaded {new_step.namespace()}. Memory usage now {round(new_memory,2)} MB. Increased by {round(new_memory-prev_memory,2)}MB, {round(mem_increase,2)}%"
        )

    return steps


def calc_doc_size(doc: Document):
    return sum(len(section.text) for section in doc.sections)


def batch_metrics(docs: List[Document]):
    lengths = []
    ent_count = []
    for doc in docs:
        lengths.append(calc_doc_size(doc))
        ent_count.append(len(doc.get_entities()))
    return {
        "max_length": max(lengths),
        "mean_length": float(sum(lengths)) / float(len(docs)),
        "max_ents": max(ent_count),
        "mean_ents": float(sum(ent_count)) / float(len(ent_count)),
    }


class FailedDocsHandler(Protocol):
    """Handle failed docs."""

    def __call__(self, step_docs_map: Dict[str, List[Document]]):
        """
        :param step_docs_map: a dict of step namespace and the docs that failed for it
        :return:
        """
        raise NotImplementedError


class FailedDocsLogHandler(FailedDocsHandler):
    """Log failed docs as warnings."""

    def __call__(self, step_docs_map: Dict[str, List[Document]]):
        for step_namespace, docs in step_docs_map.items():
            for doc in docs:
                error_message = doc.metadata.get(PROCESSING_EXCEPTION, None)
                if error_message is not None:
                    logger.warning(
                        f"processing failed for step {step_namespace}, doc: {doc.idx}, {error_message} "
                    )
                else:
                    logger.warning(
                        f"processing failed for step {step_namespace}, doc: {doc.idx}, No error mesasge found in doc metadata "
                    )


class FailedDocsFileHandler(FailedDocsHandler):
    """Log failed docs to a directory along with the relevant exception."""

    def __init__(self, log_dir: PathLike):
        self.log_dir = as_path(log_dir)

    def __call__(self, step_docs_map: Dict[str, List[Document]]):
        for step_namespace, docs in step_docs_map.items():
            step_logging_dir = self.log_dir.joinpath(step_namespace)
            step_logging_dir.mkdir(parents=True, exist_ok=True)

            for doc in docs:
                serialisable_doc = doc.json()
                doc_id = doc.idx
                doc_path = step_logging_dir.joinpath(doc_id + ".json")
                doc_error_path = step_logging_dir.joinpath(doc_id + "_error.txt")
                with open(doc_path, "w") as f:
                    f.write(serialisable_doc)
                with open(doc_error_path, "w") as f:
                    error_message = doc.metadata.get(PROCESSING_EXCEPTION)
                    if error_message is not None:
                        f.write(error_message)
                    else:
                        logger.warning(
                            f"No error message found for doc: {doc}. Cannot write exception to {doc_error_path}"
                        )


class PipelineValueError(ValueError):
    """An Exception that gets raised if arguments passed to a Pipeline are inappropriate.

    This is particularly useful to allow handling just these errors - e.g. in Kazu's web API,
    where we can then highlight that there is a problem with the data passed to the API, rather
    than an internal error to Kazu.
    """

    pass


class Pipeline:
    def __init__(
        self,
        steps: List[Step],
        failure_handler: Optional[List[FailedDocsHandler]] = None,
        profile_steps_dir: Optional[str] = None,
        skip_doc_len: Optional[int] = 200000,
        step_groups: Optional[Dict[str, Collection[str]]] = None,
    ):
        """A basic pipeline, used to help run a series of steps.

        :param steps: list of steps to run
        :param failure_handler: optional list of handlers to process failed docs
        :param profile_steps_dir: profile throughout of each step with tensorboard. path to log dir
        :param skip_doc_len: a maximum length for documents (in characters), above which they will
            be skipped. Extremely long inputs can be symptomatic of very strange text which can
            result in errors and excessive memory usage.
        :param step_groups: groups of steps to make available for running together as a
            convenience. The keys are names of groups to create, and values are the namespaces of
            steps. Order of running steps is still taken from the steps argument, not the order
            within each group. To customize step running order, you can instead use the
            ``step_namespaces`` parameter of :meth:`__call__`\\ .
        """
        self.skip_doc_len = skip_doc_len
        self.failure_handlers = failure_handler
        self.steps = steps
        self._namespace_to_step = {step.namespace(): step for step in steps}
        # documents that failed to process - a dict of [<step namespace>:List[failed docs]]
        self.failed_docs: Dict[str, List[Document]] = {}
        # performance tracking
        self.init_time = datetime.now().strftime("%m_%d_%Y_%H_%M")
        logger.info(f"pipeline initialised: {self.init_time}")
        self.profile_steps_dir = profile_steps_dir
        if profile_steps_dir:
            try:
                from torch.utils.tensorboard import SummaryWriter
            except ImportError as e:
                raise ImportError(
                    "Profiling the pipeline requires tensorboard to be installed.\n"
                    "Tensorboard is included in Kazu's dev dependencies, so this will work"
                    " with 'pip install kazu[dev]', but you could also just do 'pip install"
                    " tensorboard' if you don't want the other dev dependencies."
                ) from e

            idx_this_process = uuid.uuid1().hex
            dir_and_time = f"{profile_steps_dir}_{self.init_time}_{idx_this_process}"
            logger.info(f"profiling configured. log dir is {dir_and_time}")
            self.summary_writer: Optional[SummaryWriter] = SummaryWriter(log_dir=dir_and_time)
            self.call_count = 0
        else:
            logger.info("profiling not configured")
            self.summary_writer = None

        self.step_groups: Optional[Dict[str, List[Step]]]
        if step_groups is None:
            self.step_groups = None
        else:
            self.step_groups = {group_name: [] for group_name in step_groups}
            for step_name, step in self._namespace_to_step.items():
                for group_name, group in step_groups.items():
                    if step_name in group:
                        self.step_groups[group_name].append(step)

    def prefilter_docs(self, docs: List[Document]):
        docs_to_process = []
        for doc in docs:
            doc_size = calc_doc_size(doc)
            if doc_size >= self.skip_doc_len:
                doc.metadata[
                    PROCESSING_EXCEPTION
                ] = f"document too long: {doc_size}. max:{self.skip_doc_len}"
                logger.warning(f"skipping doc: {doc.idx}: reason: too long")
            else:
                docs_to_process.append(doc)
        return docs_to_process

    def __call__(
        self,
        docs: List[Document],
        step_namespaces: Optional[Iterable[str]] = None,
        step_group: Optional[str] = None,
    ) -> List[Document]:
        """Run the pipeline.

        :param docs: Docs to process
        :param step_namespaces: The namespaces of the steps to use in processing. Default behaviour
            is to use all steps on the pipeline in the order given when creating the pipeline.
            This parameter gives the flexibility to sometimes only run some of the steps.
        :param step_group: One of the pipeline's configured step_groups to run. This is just a
            convenience over needing to specify common groups of step_namespaces. Note that passing
            both step_group and step_namespaces is incohorent and a :exc:`ValueError` will be
            raised.
        :return: processed docs
        """
        docs_to_process = self.prefilter_docs(docs)
        step_times = {}
        batch_start = time.time()

        steps_to_run: Iterable[Step]
        if step_namespaces is not None:
            if step_group is not None:
                raise PipelineValueError(
                    "Passing both step_namespaces and step_group to Pipeline.__call__ is incomptable."
                    " Only one may be passed in a single call."
                )

            steps_to_run = []
            nonexistent_steps = []
            for namespace in step_namespaces:
                maybe_step = self._namespace_to_step.get(namespace)
                if maybe_step is None:
                    nonexistent_steps.append(namespace)
                else:
                    steps_to_run.append(maybe_step)

            if len(nonexistent_steps) >= 1:
                raise PipelineValueError(
                    "The following steps do not exist in the pipeline:\n%s"
                    "The valid steps for this pipeline are:\n%s"
                    % (nonexistent_steps, list(self._namespace_to_step))
                )

        elif step_group is not None:
            if self.step_groups is None:
                raise PipelineValueError(
                    "This pipeline does not have any step groups configured, so cannot run the"
                    " requested step_group %s" % step_group
                )

            # we can't assign directly to steps_to_run, because the type is Optional[List[Step]]
            # rather than the Iterable[Step] declared above (and required below).
            steps_or_none = self.step_groups.get(step_group)
            if steps_or_none is None:
                raise PipelineValueError(
                    "%s is not a valid step_group for this pipeline. Available step_groups:\n%s"
                    % (step_group, self.step_groups)
                )
            else:
                steps_to_run = steps_or_none

        else:
            steps_to_run = self.steps

        for step in steps_to_run:
            start = time.time()
            _processed_docs, failed_docs = step(docs_to_process)
            step_times[step.namespace()] = round(time.time() - start, 4)
            self.update_failed_docs(step, failed_docs)
        batch_time = round(time.time() - batch_start, 4)
        if self.profile_steps_dir:
            self.profile(step_times, batch_time, batch_metrics(docs))
        self.reset()

        return docs

    def profile(self, step_times: Dict, batch_time: float, batch_metrics_dict: Dict):
        if self.summary_writer is not None:
            self.summary_writer.add_scalars(
                main_tag="batch_metrics",
                tag_scalar_dict=batch_metrics_dict,
                global_step=self.call_count,
            )
            self.summary_writer.add_scalars(
                main_tag="all_steps", tag_scalar_dict=step_times, global_step=self.call_count
            )
            self.summary_writer.add_scalars(
                main_tag="batch_time",
                tag_scalar_dict={"batch": batch_time},
                global_step=self.call_count,
            )
            self.summary_writer.add_scalars(
                main_tag="memory",
                tag_scalar_dict={"MB": psutil.Process(os.getpid()).memory_info().rss / 1024**2},
                global_step=self.call_count,
            )
            for step_name, step_time in step_times.items():
                self.summary_writer.add_scalars(
                    main_tag=step_name,
                    tag_scalar_dict={"time": step_time},
                    global_step=self.call_count,
                )
            self.call_count += 1

    def update_failed_docs(self, step: Step, failed_docs: List[Document]):
        if self.failure_handlers is not None:
            self.failed_docs[step.namespace()] = failed_docs

    def reset(self):
        if self.failure_handlers is not None:
            for handler in self.failure_handlers:
                handler(self.failed_docs)
        self.failed_docs = {}



import logging
import statistics
import time
from typing import List

logger = logging.getLogger(__name__)


class Stopwatch:
    def __init__(self, call_count_interval: int = 1):
        """
        a stop watch! will produce a logging message from the Stopwatch.message().

        :param call_count_interval: when message()  is called, only log if self.call_count %
                                    self.call_count_interval == 0
        """
        self.call_count_interval = call_count_interval
        self.call_count = 0
        self.times: List[float] = []

    def start(self):
        self.mark = time.time()

    def message(self, message: str):
        """
        helper function for benchmarking in logs

        :param start:
        :return:
        """

        end = time.time()
        self.call_count += 1

        time_elapsed = round(end - self.mark, 2)
        self.last_average = None
        self.times.append(time_elapsed)
        self.mark = end

        if self.call_count % self.call_count_interval == 0:
            if self.last_average is not None:
                new_average = statistics.mean(self.times)
                result = round((new_average + self.last_average) / 2, 2)
                self.last_average = new_average
                self.times = []
                logger.info(
                    f"{message}:{time_elapsed}, average: {result}, total calls:{self.call_count}, "
                )
            else:
                self.last_average = statistics.mean(self.times)
                self.times = []



from functools import cache

import stanza
from stanza.pipeline.core import DownloadMethod

from kazu.utils.utils import PathLike


@cache
def simple_stanza_init(path: PathLike, download: bool, use_gpu: bool) -> stanza.Pipeline:
    if download:
        stanza.download(lang="en", package="genia", model_dir=path)
    stanza_pipeline = stanza.Pipeline(
        lang="en",
        model_dir=str(path),
        package=None,
        processors={"tokenize": "genia"},
        use_gpu=use_gpu,
        download_method=DownloadMethod.REUSE_RESOURCES,
    )
    return stanza_pipeline
